
; Pres3D
; (w) 2010 by Volker Schmid (v.schmid@inspirant.de)
;
; Licence: LGPL (http://www.gnu.org/licenses/lgpl.html)
;          english information: http://en.wikipedia.org/wiki/GNU_Lesser_General_Public_License
;          german information:  http://de.wikipedia.org/wiki/LGPL

#Version = "1.03"

Enumeration
  #MenuOpen
  #MenuExit
  #MenuAbout
EndEnumeration


UsePNGImageDecoder()
UsePNGImageEncoder()
UseJPEGImageDecoder()

IncludeFile "Supersprite3D.pbi" ; this uses the SuperSprite3D include

; effect settings structure
Structure structEffectParameters
  Type.s
  Direction.s
  Rotation.s
  Misc.s
EndStructure

; objects definition structure
Structure structObject
  Type.b ; 0=char 1=image 2=?
  
  SpriteID.SuperSprite3D ; SuperSprite3D ID
  
  SpriteImageID.l ; ID of the loaded sprite (to free up and copy)
  
  Transform.b ; #True, if transforming is in progress
  
  PosX.f ; desired position X
  PosY.f ; desired position Y
  PosZ.f ; desired position Z
  
  ActualPosX.f ; actual position X
  ActualPosY.f ; actual position Y
  ActualPosZ.f ; actual position Z
  
  AngleX.f ; desired angle X
  AngleY.f ; desired angle Y
  AngleZ.f ; desired angle Z
  
  ActualAngleX.f ; actual angle X
  ActualAngleY.f ; actual angle Y
  ActualAngleZ.f ; actual angle Z
  
  Transparency.l       ; desired Transparency
  ActualTransparency.l ; actual Transparency
  
  Val0.f
  Val1.f
  Val2.f
  Val3.f
  Val4.f ; free variables for effect routines
  Val5.f
  Val6.f
  Val7.f
  Val8.f
  Val9.f
  
  EffectIn.structEffectParameters  ; effect ID for ingoing
  EffectOut.structEffectParameters ; effect ID for outgoing
  
  GroupID.l    ; ID in my group (1 to GroupMax.l)
  GroupMax.l   ; max ID in my group
  
  Delay.l      ; delay in milliseconds
  
  ImageSplitSize.l ; split image into tiles with this size
  
EndStructure

; program settings structure
Structure structSettings
  ScreenFull.b    ; #True or #False
  ScreenWidth.l
  ScreenHeight.l
  BackgroundColor.l 
  Smoothing.b     ; #True or #False
EndStructure

#BufferImages = 6000 ; maximum number of elements per page
#XML          = 0    ; XML-ID
#InitialZ     = 578  ; initial Z-Position of (currently) all new elements

Global XML_CurrentPageNode.l        ; holds the next XML node to read the next page from
Global XML_FilePath.s               ; path (incl. last \) of the currently loaded xml file
Global XML_CurrentXMLFile.s         ; the currently loaded xml file
Global XML_Settings.structSettings  ; global settings of the currently loaded XML
Global XML_MainNode.l               ; global XML main node

Global IsFullscreen.b               ; #True if Fullscreen-Mode is active, #False if not

Global Dim InObjects.structObject(#BufferImages)
Global InObjectCount.l = 0 ; counter for ingoing objects
Global Dim OutObjects.structObject(#BufferImages)
Global OutObjectCount.l = 0 ; counter for outgoing objects

CompilerIf #PB_Compiler_OS = #PB_OS_Windows
  ; allow Sprite3D view from behind (windows-only!!!!!!)
  Define pd3d.IDirect3DDevice9
CompilerEndIf

InitSprite() : InitSuperSprite3D() : InitKeyboard() : InitMouse()

; ---------------------------------------------------------------
; Effects
; ---------------------------------------------------------------
IncludeFile "Effects.pbi"

; ---------------------------------------------------------------
; Procedures
; ---------------------------------------------------------------
  
; open a screen for displaying
Procedure OpenDisplayScreen()
  
  If IsWindow(0) <> 0
    ; close previous opened window
    CloseScreen()
    CloseWindow(0)
  EndIf
  
  ScreenRet.l = OpenScreen(XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, 32, "Pres3D")
  
  IsFullscreen.b = #True
  
  ReleaseMouse(0)
  
  If XML_Settings\Smoothing = #True
    Sprite3DQuality(#PB_Sprite3D_BilinearFiltering)
  Else
    Sprite3DQuality(#PB_Sprite3D_NoFiltering)
  EndIf
  
  ProcedureReturn ScreenRet.l
  
EndProcedure

; open a window for displaying
Procedure OpenDisplayWindow()
  
  If IsFullscreen.b = #True
    CloseScreen()
  EndIf
  
  If Not OpenWindow(0, 0, 0, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, "Pres3D V" + #Version, #PB_Window_ScreenCentered | #PB_Window_SystemMenu)
    MessageRequester("error", "Can not open screen! Please check your values...", #MB_ICONERROR)
    End
  EndIf
  
  CreateMenu(0, WindowID(0))
  MenuTitle("File")
  MenuItem(#MenuOpen, "Open")
  MenuBar()
  MenuItem(#MenuExit, "Exit")
  MenuTitle("Help")
  MenuItem(#MenuAbout, "About")

  ScreenRet.l = OpenWindowedScreen(WindowID(0), 0, 0, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, 0, 0, 0)
  
  IsFullscreen.b = #False
  
  ReleaseMouse(1)
  
  If XML_Settings\Smoothing = #True
    Sprite3DQuality(#PB_Sprite3D_BilinearFiltering)
  Else
    Sprite3DQuality(#PB_Sprite3D_NoFiltering)
  EndIf

  ProcedureReturn ScreenRet.l
  
EndProcedure

      

; ImageToSprite function copy's an image to a sprite without the need to safe the
; image as png before. Written by STARGTE (german PureBasic Forum).
; Adapted and reformated by V. Schmid (Oct 2010).
Procedure ImageToSprite(Sprite.l, Image.l, Mode.l = #False)
  Protected Row.l
  Protected ImageHeight.l = ImageHeight(Image.l)
  Protected ImageWidth.l  = ImageWidth(Image.l)
  Protected *ReadBuffer.l,  ReadPitch.l,  ReadFormat.l
  Protected *WriteBuffer.l, WritePitch.l, WriteFormat.l 
  StartDrawing(ImageOutput(Image.l))
    *ReadBuffer.l = DrawingBuffer()
    ReadPitch.l   = DrawingBufferPitch()
    ReadFormat.l  = DrawingBufferPixelFormat()
  StopDrawing()
  If Sprite = #PB_Any
    Sprite.l = CreateSprite(#PB_Any, ImageWidth.l, ImageHeight.l, Mode.l)
  Else
    CreateSprite(Sprite.l, ImageWidth.l, ImageHeight.l, Mode.l)
  EndIf
  StartDrawing(SpriteOutput(Sprite.l))
    *WriteBuffer.l = DrawingBuffer()
    WritePitch.l   = DrawingBufferPitch()
    WriteFormat.l  = DrawingBufferPixelFormat()
    If ReadFormat.l & #PB_PixelFormat_ReversedY XOr WriteFormat.l & #PB_PixelFormat_ReversedY
      For Row.l = 0 To ImageHeight.l - 1
        CopyMemory(*ReadBuffer.l + (ImageHeight.l - 1 - Row.l) * ReadPitch.l, *WriteBuffer.l + Row.l * WritePitch.l, ImageWidth.l * SizeOf(Long))
      Next
    Else
      For Row.l = 0 To ImageHeight.l - 1
        CopyMemory(*ReadBuffer.l + Row.l * ReadPitch.l, *WriteBuffer.l + Row.l * WritePitch.l, ImageWidth.l * SizeOf(Long))
      Next
    EndIf
  StopDrawing()
  ProcedureReturn Sprite.l
EndProcedure

; Returns the width (pixel) of a char respecting it's surrounding chars...
; IMPORTANT: This needs a current StartDrawing() context and a defined font!
Procedure.l GetCharWidth(Text.s, CharPos.l)
  If CharPos.l < 2
    ; first char
    Char1.s = Left(Text.s, 2) ; the first two
    Char2.s = Mid(Text.s, 2, 1); without the needed
    ProcedureReturn TextWidth(Char1.s) - TextWidth(Char2.s)
  EndIf
  If CharPos.l > Len(Text.s) - 1
    ; last char
    Char1.s = Right(Text.s, 2) ; the last two
    Char2.s = Right(Text.s, 1); without the needed
    ProcedureReturn TextWidth(Char1.s) - TextWidth(Char2.s)
  EndIf
  If CharPos.l > 1 And CharPos.l < Len(Text.s)
    ; all other chars
    Char1.s = Mid(Text.s, CharPos.l - 1, 3) ; all three
    Char2.s = Left(Char1.s, 1) + Right(Char1.s, 1); without the needed
    ProcedureReturn TextWidth(Char1.s) - TextWidth(Char2.s)
  EndIf
EndProcedure

; return a free SpriteID that is currently and definitely not used as sprite
Procedure.l GetFreeSpriteID()
  Protected SpriteID.l

  Repeat
    SpriteID.l = Random(50000)
  Until IsSprite(SpriteID.l) = 0
  ProcedureReturn SpriteID.l
EndProcedure

; insert a new group ob objects containing text to the ingoing objects
Procedure.l MakeTextGroup(Text.s, FontName.s, Size.f, PosX.l, PosY.l, Color.l, Style.l, Delay.l, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  LoadFont(0, FontName.s, Size.f, Style.l) ; load Arial font in desired size
  
  BoxSize.l = 16
  
  If Size.f > 15
    BoxSize.l = 32
  EndIf
  If Size.f > 20
    BoxSize.l = 64
  EndIf
  If Size.f > 40
    BoxSize.l = 128
  EndIf
  
  ColR.l = Red(Color.l)
  ColG.l = Green(Color.l)
  ColB.l = Blue(Color.l)
  
  ; generate char sprites
  
  LeftPos.l = 0; init
  For x.l = 1 To Len(Text.s)
    
    InObjectCount.l = InObjectCount.l + 1
    
    ; create sprite
    CreateImage(0, BoxSize.l, BoxSize.l, #PB_Image_Transparent  | 32)
    StartDrawing(ImageOutput(0))
    Box(0, 0, BoxSize.l, BoxSize.l, RGBA(255,255,255,0))
    
    ; set font
    DrawingFont(FontID(0))
    
    MaxWi.l = TextWidth("_")
    MaxHe.l = TextHeight("|")
    
    CenterX.l = - TextWidth(Text.s) / 2
    CenterY.l = - TextHeight(Text.s) / 2
    
    Char.s = Mid(Text.s, x.l, 1)
    Wi.l = GetCharWidth(Text.s, x.l)
    
    ; create object structure
    With InObjects(InObjectCount.l)
      \Type         = 0; text
      \PosX         = PosX.l + LeftPos.l + CenterX.l + MaxWi.l / 2
      \PosY         = PosY.l; + CenterY.l
      \PosZ         = #InitialZ
      \ActualPosX   = \PosX
      \ActualPosY   = \PosY
      \ActualPosZ   = \PosZ
      
      \Transparency        = 255
      \ActualTransparency  = \Transparency
      
      \EffectIn\Type       = *EffectIn\Type
      \EffectIn\Direction  = *EffectIn\Direction
      \EffectIn\Rotation   = *EffectIn\Rotation
      \EffectIn\Misc       = *EffectIn\Misc
      
      \EffectOut\Type      = *EffectOut\Type
      \EffectOut\Direction = *EffectOut\Direction
      \EffectOut\Rotation  = *EffectOut\Rotation
      \EffectOut\Misc      = *EffectOut\Misc
      
      \GroupID      = x.l
      \GroupMax     = Len(Text.s)
      
      \Transform.b  = #True
      
      ; init
      \ActualAngleX = 0
      \ActualAngleY = 0
      \ActualAngleZ = 0
      \AngleX       = 0
      \AngleY       = 0
      \AngleZ       = 0
      
      \Delay        = Delay.l
      
    EndWith
    LeftPos.l = LeftPos.l + Wi.l
    
    ; draw char to image
    DrawingMode(#PB_2DDrawing_AllChannels)
    ; DrawingMode(#PB_2DDrawing_AllChannels | #PB_2DDrawing_Gradient)
    ; CircularGradient(BoxSize.l/2, BoxSize.l/2, BoxSize.l/3) : GradientColor(0.0,RGBA(0,255,0,255)) : GradientColor(1.0,RGBA(0,0,0, 255))
    DrawText((BoxSize.l - MaxWi.l) / 2, (BoxSize.l - MaxHe.l) / 2, Char.s, RGBA(ColR.l,ColG.l,ColB.l,255), RGBA(0,0,0,0))
    StopDrawing()
    
    ; load ingoing image as sprite
    SpriteID.l = GetFreeSpriteID()
    ImageToSprite(SpriteID.l, 0, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
    InObjects(InObjectCount.l)\SpriteImageID = SpriteID.l
    
    CreateSuperSprite3D(InObjects(InObjectCount.l)\SpriteID, SpriteID.l, BoxSize.l, BoxSize.l)
    
    EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount.l, x.l, Len(Text.s))
    
    FreeImage(0)
    
  Next
  
EndProcedure

; insert a new group ob objects containing text to the ingoing objects
Procedure.l MakeTextBlock(Text.s, FontName.s, Size.f, PosX.l, PosY.l, Color.l, Style.l, Delay.l, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  LoadFont(0, FontName.s, Size.f, Style.l) ; load Arial font in desired size
  
  InObjectCount.l = InObjectCount.l + 1
  
  ; create test-image (for getting the needed width)
  CreateImage(0, 512, 512, #PB_Image_Transparent  | 32)
  StartDrawing(ImageOutput(0))
  ; set font
  DrawingFont(FontID(0))
  ; get needed size
  Size.f = TextWidth(Text.s)
  ; free it
  StopDrawing()
  FreeImage(0)
  
  BoxSize.l = 16
  If Size.f > 16: BoxSize.l = 32: EndIf
  If Size.f > 32: BoxSize.l = 64: EndIf
  If Size.f > 64: BoxSize.l = 128: EndIf
  If Size.f > 128: BoxSize.l = 256: EndIf
  If Size.f > 256: BoxSize.l = 512: EndIf
  If Size.f > 512
    BoxSize.l = Round(Size.f / 10, #PB_Round_Up) * 10
    ; Debug "Box-Size (!): " + Str(BoxSize.l)
  EndIf
  
  ; create object structure
  With InObjects(InObjectCount.l)
    \Type         = 0; text
    \PosX         = PosX.l; - Size.f / 2
    \PosY         = PosY.l; - Size.f / 2
    \PosZ         = #InitialZ
    \ActualPosX   = \PosX
    \ActualPosY   = \PosY
    \ActualPosZ   = \PosZ
    
    \Transparency        = 255
    \ActualTransparency  = \Transparency
    
    \EffectIn\Type       = *EffectIn\Type
    \EffectIn\Direction  = *EffectIn\Direction
    \EffectIn\Rotation   = *EffectIn\Rotation
    \EffectIn\Misc       = *EffectIn\Misc
    
    \EffectOut\Type      = *EffectOut\Type
    \EffectOut\Direction = *EffectOut\Direction
    \EffectOut\Rotation  = *EffectOut\Rotation
    \EffectOut\Misc      = *EffectOut\Misc
    
    \GroupID      = 1
    \GroupMax     = 1
    
    \Transform.b  = #True
    
    ; init
    \ActualAngleX = 0
    \ActualAngleY = 0
    \ActualAngleZ = 0
    \AngleX       = 0
    \AngleY       = 0
    \AngleZ       = 0
    
    \Delay        = Delay.l
    
  EndWith
  
  ColR.l = Red(Color.l)
  ColG.l = Green(Color.l)
  ColB.l = Blue(Color.l)
  
  ; create sprite
  CreateImage(0, BoxSize.l, BoxSize.l, #PB_Image_Transparent  | 32)
  StartDrawing(ImageOutput(0))
  Box(0, 0, BoxSize.l, BoxSize.l, RGBA(255,255,255,0))
  
  ; set font
  DrawingFont(FontID(0))
  
  CenterX.l = (BoxSize.l - TextWidth(Text.s)) / 2
  CenterY.l = (BoxSize.l - TextHeight(Text.s)) / 2
    
  ; draw char to image
  DrawingMode(#PB_2DDrawing_AllChannels)
  ; Box(0, 0, BoxSize.l, BoxSize.l, RGBA(55,55,75,255))
  DrawText(CenterX.l, CenterY.l, Text.s, RGBA(ColR.l,ColG.l,ColB.l,255), RGBA(0,0,0,0))
  StopDrawing()
  
  ; load ingoing image as sprite
  SpriteID.l = GetFreeSpriteID()
  ImageToSprite(SpriteID.l, 0, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
  InObjects(InObjectCount.l)\SpriteImageID = SpriteID.l
  
  CreateSuperSprite3D(InObjects(InObjectCount.l)\SpriteID, SpriteID.l, BoxSize.l, BoxSize.l)
  
  EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount.l, 1, 1)
  
  FreeImage(0)
EndProcedure

; insert a new image to the ingoing objects
Procedure.l MakeImage(Filename.s, PosX.l, PosY.l, Width.l, Height.l, Delay.l, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  If LoadImage(0, Filename.s) = 0
    ProcedureReturn 0
  EndIf
  
  InObjectCount.l = InObjectCount.l + 1
    
  
  If Width.l < 1 : Width.l  = ImageWidth(0) : EndIf
  If Height.l < 1: Height.l = ImageHeight(0): EndIf
  
  If ImageWidth(0) > ImageHeight(0)
    ; landscape
    Widest.l = Width.l
  Else
    ; portrait
    Widest.l = Height.l
  EndIf
  For x.l = 4 To 18
    If Pow(2, x.l) >= Widest.l
      BoxSize.l = Pow(2, x.l)
      Break
    EndIf
  Next
  
  ; create object structure
  With InObjects(InObjectCount.l)
    \Type         = 1; image
    \PosX         = PosX.l ;- (BoxSize.l / 2)
    \PosY         = PosY.l ;- (BoxSize.l / 2)
    \PosZ         = #InitialZ
    \ActualPosX   = \PosX
    \ActualPosY   = \PosY
    \ActualPosZ   = \PosZ
    
    \Transparency        = 255
    \ActualTransparency  = \Transparency
    
    \EffectIn\Type       = *EffectIn\Type
    \EffectIn\Direction  = *EffectIn\Direction
    \EffectIn\Rotation   = *EffectIn\Rotation
    \EffectIn\Misc       = *EffectIn\Misc
    
    \EffectOut\Type      = *EffectOut\Type
    \EffectOut\Direction = *EffectOut\Direction
    \EffectOut\Rotation  = *EffectOut\Rotation
    \EffectOut\Misc      = *EffectOut\Misc
    
    \Transform.b  = #True
    
    ; init
    \ActualAngleX = 0
    \ActualAngleY = 0
    \ActualAngleZ = 0
    \AngleX       = 0
    \AngleY       = 0
    \AngleZ       = 0
    
    \GroupID      = 1
    \GroupMax     = 1
    
    \Delay        = Delay.l
    
  EndWith
  
  CreateImage(1, BoxSize.l, BoxSize.l, 32 | #PB_Image_Transparent)
  
  StartDrawing(ImageOutput(1))
    DrawingMode(#PB_2DDrawing_AlphaBlend)
    ResizeImage(0, Width.l, Height.l, #PB_Image_Smooth)
    DrawImage(ImageID(0), (BoxSize.l - Width.l) / 2, (BoxSize.l - Height.l) / 2, Width.l, Height.l)
    ; Box(0,0,BoxSize.l, BoxSize.l, RGBA(200,200,200,200)) ; for debug only
  StopDrawing()
  
  SpriteID.l = GetFreeSpriteID()
  ImageToSprite(SpriteID.l, 1, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
  InObjects(InObjectCount.l)\SpriteImageID = SpriteID.l
  
  CreateSuperSprite3D(InObjects(InObjectCount.l)\SpriteID, SpriteID.l, SpriteWidth(SpriteID.l), SpriteHeight(SpriteID.l))
  
  EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount.l, 1, 1)
  
  FreeImage(1)
  FreeImage(0)
  
EndProcedure

; insert a new image to the ingoing objects. This one is split into multiple smaller quadratic images.
Procedure.l MakeImageSplit(Filename.s, PosX.l, PosY.l, Width.l, Height.l, Delay.l, SplitSize.l, *EffectIn.structEffectParameters, *EffectOut.structEffectParameters)
  
  If LoadImage(0, Filename.s) = 0 ; load original image
    ProcedureReturn 0
  EndIf
  
  If Width.l < 1 : Width.l  = ImageWidth(0) : EndIf
  If Height.l < 1: Height.l = ImageHeight(0): EndIf
  
  If ImageWidth(0) > ImageHeight(0)
    ; landscape
    Widest.l = Width.l
  Else
    ; portrait
    Widest.l = Height.l
  EndIf
  For x.l = 4 To 18
    If Pow(2, x.l) >= Widest.l
      BoxSize.l = Pow(2, x.l)
      Break
    EndIf
  Next
  
  ResizeImage(0, Width.l, Height.l, #PB_Image_Smooth)
  
  CreateImage(1, BoxSize.l, BoxSize.l, 32 | #PB_Image_Transparent) ; create scaled image
  
  StartDrawing(ImageOutput(1)); start drawing to scaled image
  DrawingMode(#PB_2DDrawing_AlphaBlend)
  ; Box(1,1,BoxSize.l, BoxSize.l, RGBA(100,100,100,100))
  DrawImage(ImageID(0), (BoxSize.l - Width.l) / 2, (BoxSize.l - Height.l) / 2, Width.l, Height.l); insert original image to this block
  StopDrawing()
  
  ; in image 1 is the correct and complete image to display
  
  Parts.l = BoxSize.l / SplitSize.l
  
  For x.l = 0 To Parts.l - 1
    For y.l = 0 To Parts.l - 1
      
      InObjectCount.l = InObjectCount.l + 1
      
      ; create object structure
      With InObjects(InObjectCount.l)
        \Type         = 1; image
        \PosX         = PosX.l + (x.l * SplitSize.l) - (BoxSize.l / 2) + Splitsize.l / 2
        \PosY         = PosY.l + (y.l * SplitSize.l) - (BoxSize.l / 2) + SplitSize.l / 2
        \PosZ         = #InitialZ
        \ActualPosX   = \PosX
        \ActualPosY   = \PosY
        \ActualPosZ   = \PosZ
        
        \Transparency        = 255
        \ActualTransparency  = \Transparency
        
        \EffectIn\Type       = *EffectIn\Type
        \EffectIn\Direction  = *EffectIn\Direction
        \EffectIn\Rotation   = *EffectIn\Rotation
        \EffectIn\Misc       = *EffectIn\Misc
        
        \EffectOut\Type      = *EffectOut\Type
        \EffectOut\Direction = *EffectOut\Direction
        \EffectOut\Rotation  = *EffectOut\Rotation
        \EffectOut\Misc      = *EffectOut\Misc
        
        \Transform.b  = #True
        
        ; init
        \ActualAngleX = 0
        \ActualAngleY = 0
        \ActualAngleZ = 0
        \AngleX       = 0
        \AngleY       = 0
        \AngleZ       = 0
        
        \GroupID      = y.l * 3
        \GroupMax     = Parts.l * 3
        
        \Delay        = Delay.l
        
      EndWith
      
      StartDrawing(ImageOutput(1)); start drawing to scaled image
      
      GrabDrawingImage(2, x.l * SplitSize.l, y.l * SplitSize.l, SplitSize.l, SplitSize.l)
      
      StopDrawing()
      
      SpriteID.l = GetFreeSpriteID()
      ImageToSprite(SpriteID.l, 2, #PB_Sprite_Texture|#PB_Sprite_AlphaBlending)
      
      InObjects(InObjectCount.l)\SpriteImageID = SpriteID.l
      
      CreateSuperSprite3D(InObjects(InObjectCount.l)\SpriteID, SpriteID.l, SpriteWidth(SpriteID.l), SpriteHeight(SpriteID.l))
      
      EffectInitIn(*EffectIn, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, InObjectCount.l, y.l * 3, Parts.l * 3)
      
      FreeImage(2)
      
    Next ; y.l
  Next ; x.l
  
  FreeImage(1)  ; free scaled image
  
  FreeImage(0)  ; free initial image
  
EndProcedure

; Make all current ingoing objects to outgoing objects (will start to fade out)
Procedure ActivateOutgoing()
  
  ; clean all current outgoing objects (their animation has been finished)
  For x.l = 1 To OutObjectCount.l
    If OutObjects(x.l)\PosX > 0
      FreeSprite3D(OutObjects(x.l)\SpriteImageID)
      FreeSprite(OutObjects(x.l)\SpriteImageID)
      ; not shure, if it needs some kind of killing the assigned SuperSprite3D instance here?
    EndIf
  Next
  
  Dim OutObjects.structObject(#BufferImages) ; clean array
  
  ; copy all currently displayed objects (InObjects) to outgoing
  For x.l = 1 To InObjectCount.l
    ; copy
    OutObjects.structObject(x.l) = InObjects.structObject(x.l)
    
    ; start their outgoing effect
    *Out.structEffectParameters = OutObjects(x.l)\EffectOut
    EffectInitOut(*Out, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, x.l, OutObjects(x.l)\GroupID, OutObjects(x.l)\GroupMax)
  Next ; x.l
  
  OutObjectCount.l = InObjectCount.l
  
  ; clean old ingoing objects
  Dim InObjects.structObject(#BufferImages) ; clean array for ingoing objects
  InObjectCount.l = 0
EndProcedure

; ---------------------------------------------------------------
; XML handling
; ---------------------------------------------------------------

; get the effect (effectin or effectout) parameters in the given XML node
Procedure.l GetEffectNode(*XMLNode, EffectName.s)
  Protected *Result.structEffectParameters = AllocateMemory(SizeOf(structEffectParameters))
  
  EffectName.s = LCase(EffectName.s)
  
  Count.l = XMLChildCount(*XMLNode)
  
  For x.l = 1 To Count.l
    *ChildNode = ChildXMLNode(*XMLNode, x.l)
    Name.s = LCase(GetXMLNodeName(*ChildNode))
    If Name.s = EffectName.s
      ; effect found
      *Result\Type      = GetXMLAttribute(*ChildNode, "type")
      *Result\Direction = GetXMLAttribute(*ChildNode, "direction")
      *Result\Rotation  = GetXMLAttribute(*ChildNode, "rotation")
      *Result\Misc      = GetXMLAttribute(*ChildNode, "misc")
      ProcedureReturn *Result
    EndIf
  Next
  
EndProcedure

; get all text and images from the current page (*XMLNode) to the ingoing objects
Procedure InsertXMLElements(*XMLNode)
  *ChildNode = ChildXMLNode(*XMLNode)
  While *ChildNode <> 0
    If *ChildNode <> 0
      ; node found
      NodeName.s = GetXMLNodeName(*ChildNode)
      ; Debug "Node: " + NodeName.s
      
      Select NodeName.s
        Case "text"
          Caption.s   =     GetXMLAttribute(*ChildNode, "caption")
          PosX.l      = Val(GetXMLAttribute(*ChildNode, "posx"))
          PosY.l      = Val(GetXMLAttribute(*ChildNode, "posy"))
          FontName.s  =     GetXMLAttribute(*ChildNode, "fontname")
          FontSize.f  = Val(GetXMLAttribute(*ChildNode, "fontsize"))
          FontStyle.l = Val(GetXMLAttribute(*ChildNode, "fontstyle"))
          Color.s     =     GetXMLAttribute(*ChildNode, "color")
          Delay.l     = Val(GetXMLAttribute(*ChildNode, "delay"))
          
          *EffectIn.structEffectParameters  = GetEffectNode(*ChildNode, "effectin")
          *EffectOut.structEffectParameters = GetEffectNode(*ChildNode, "effectout")
          
          Color.s     = ReplaceString(Color.s, "#" , ""); ignore # (html style)
          
          ColorVal.l  = RGB(Val("$"+Mid(Color.s,1,2)),Val("$"+Mid(Color.s,3,2)),Val("$"+Mid(Color.s,5,2))); parse hex color values
          
          MakeTextGroup(Caption.s, FontName.s, FontSize.f, PosX.l, PosY.l, ColorVal.l, FontStyle.l, Delay.l, *EffectIn, *EffectOut)
          
        Case "textblock"
          Caption.s   =     GetXMLAttribute(*ChildNode, "caption")
          PosX.l      = Val(GetXMLAttribute(*ChildNode, "posx"))
          PosY.l      = Val(GetXMLAttribute(*ChildNode, "posy"))
          FontName.s  =     GetXMLAttribute(*ChildNode, "fontname")
          FontSize.f  = Val(GetXMLAttribute(*ChildNode, "fontsize"))
          FontStyle.l = Val(GetXMLAttribute(*ChildNode, "fontstyle"))
          Color.s     =     GetXMLAttribute(*ChildNode, "color")
          Delay.l     = Val(GetXMLAttribute(*ChildNode, "delay"))
          
          *EffectIn.structEffectParameters  = GetEffectNode(*ChildNode, "effectin")
          *EffectOut.structEffectParameters = GetEffectNode(*ChildNode, "effectout")
          
          Color.s     = ReplaceString(Color.s, "#" , ""); ignore # (html style)
          
          ColorVal.l  = RGB(Val("$"+Mid(Color.s,1,2)),Val("$"+Mid(Color.s,3,2)),Val("$"+Mid(Color.s,5,2))); parse hex color values
          
          MakeTextBlock(Caption.s, FontName.s, FontSize.f, PosX.l, PosY.l, ColorVal.l, FontStyle.l, Delay.l, *EffectIn, *EffectOut)
          
        Case "image"
          PosX.l      = Val(GetXMLAttribute(*ChildNode, "posx"))
          PosY.l      = Val(GetXMLAttribute(*ChildNode, "posy"))
          Width.l     = Val(GetXMLAttribute(*ChildNode, "width"))
          Height.l    = Val(GetXMLAttribute(*ChildNode, "height"))
          Image.s     =     GetXMLAttribute(*ChildNode, "image")
          Delay.l     = Val(GetXMLAttribute(*ChildNode, "delay"))
          SplitSize.l = Val(GetXMLAttribute(*ChildNode, "splitsize"))
          
          *EffectIn.structEffectParameters  = GetEffectNode(*ChildNode, "effectin")
          *EffectOut.structEffectParameters = GetEffectNode(*ChildNode, "effectout")
          
          If SplitSize.l >= 4
            MakeImageSplit(Image.s, PosX.l, PosY.l, Width.l, Height.l, Delay.l, SplitSize.l, *EffectIn, *EffectOut)
          Else
            MakeImage(Image.s, PosX.l, PosY.l, Width.l, Height.l, Delay.l, *EffectIn, *EffectOut)
          EndIf
          
      EndSelect
      
    EndIf
    
    *ChildNode = NextXMLNode(*ChildNode)
  Wend
  
EndProcedure

; open the XMLFile.s for a slideshow
Procedure.b OpenSlideXML(XMLFile.s)
  
  If IsXML(#XML) <> 0
    ; cleanup current presentation
    FreeXML(#XML)
    XML_CurrentPageNode   = 0
    XML_CurrentXMLFile.s  = ""
    XML_FilePath.s        = ""
    
    InObjectCount = 0
    For i = 0 To ArraySize(InObjects())
      If IsSprite3D(InObjects(i)\SpriteID)
        FreeSprite3D(InObjects(i)\SpriteID)
      EndIf
      If IsSprite(InObjects(i)\SpriteImageID)
        FreeSprite(InObjects(i)\SpriteImageID)
      EndIf
    Next i
    
    ReDim InObjects.structObject(0)
    ReDim InObjects.structObject(#BufferImages)
    
    OutObjectCount = 0
    For i = 0 To ArraySize(OutObjects())
      If IsSprite3D(OutObjects(i)\SpriteID)
        FreeSprite3D(OutObjects(i)\SpriteID)
      EndIf
      If IsSprite(OutObjects(i)\SpriteImageID)
        FreeSprite(OutObjects(i)\SpriteImageID)
      EndIf
    Next i
    
    ReDim OutObjects.structObject(0)
    ReDim OutObjects.structObject(#BufferImages)
    
  EndIf
  
  If LoadXML(#XML, XMLFile.s) = 0
    XMLFile = OpenFileRequester("Choose a Presentation", GetPathPart(ProgramFilename()) + "slide.xml", "XML (*.xml)|*.xml", 0 )
    If LoadXML(#XML, XMLFile.s) = 0
      ProcedureReturn #False
    EndIf
  EndIf
 
  Repeat
    If XMLStatus(#XML) <> #PB_XML_Success
      res = MessageRequester("Info", "Sorry, this was no suitable slide-file. Like to open another one ?", #PB_MessageRequester_YesNo| $50)
      If res = #PB_MessageRequester_Yes       
        XMLFile = OpenFileRequester("Again", GetPathPart(ProgramFilename()), "XML (*.xml)|*.xml", 0 )
        If LoadXML(#XML, XMLFile.s) = 0
          ProcedureReturn #False
        EndIf
      Else
        MessageRequester("Oh Well", "so maybe some other time ", $50)
        ProcedureReturn #False
      EndIf
    EndIf
  Until XMLStatus(#XML) = #PB_XML_Success
  
  
  If XMLStatus(#XML) <> #PB_XML_Success
    Message.s = "Error in the XML file: "
    Message + "Message: " + XMLError(#XML) + " "
    Message + "Line: " + Str(XMLErrorLine(#XML)) + " Character: " + Str(XMLErrorPosition(#XML))
    Debug Message
    ProcedureReturn #False
  EndIf
  
  XML_MainNode = MainXMLNode(#XML)
  
  XMLName.s = GetXMLNodeName(XML_MainNode)
  If XMLName <> "pres3d"
    Debug "No valid Pres3D xml file!"
    ProcedureReturn #False
  EndIf
  
  XML_CurrentPageNode.l = ChildXMLNode(XML_MainNode)
  XML_FilePath          = GetPathPart(XMLFile.s)
  XML_CurrentXMLFile.s  = XMLFile.s
  
  ; --- read settings in current xml structure ---
  
  *ChildNode = ChildXMLNode(XML_MainNode)
  ; defaults
  XML_Settings\ScreenFull       = 0
  XML_Settings\ScreenWidth      = 1024
  XML_Settings\ScreenHeight     = 768
  XML_Settings\BackgroundColor  = RGB(255,255,255)
  ; read values (if given)
  While *ChildNode <> 0
    If *ChildNode <> 0 And XMLNodeType(*ChildNode) = #PB_XML_Normal 
      ; node found
      NodeName.s = GetXMLNodeName(*ChildNode)
      If NodeName.s = "settings"
        XML_Settings\ScreenFull   = Val(GetXMLAttribute(*ChildNode, "screenfull"))
        XML_Settings\ScreenWidth  = Val(GetXMLAttribute(*ChildNode, "screenwidth"))
        XML_Settings\ScreenHeight = Val(GetXMLAttribute(*ChildNode, "screenheight"))
        Color.s                   =     GetXMLAttribute(*ChildNode, "backgroundcolor")
        If GetXMLAttribute(*ChildNode, "smoothing") = "true" Or GetXMLAttribute(*ChildNode, "smoothing") = "1"
          XML_Settings\Smoothing  = #True
        Else
          XML_Settings\Smoothing  = #False
        EndIf
        
        Color.s = ReplaceString(Color.s, "#" , ""); ignore # (html style)  
        ColorVal.l = RGB(Val("$"+Mid(Color.s,1,2)),Val("$"+Mid(Color.s,3,2)),Val("$"+Mid(Color.s,5,2))); parse hex color values
        XML_Settings\BackgroundColor = ColorVal.l
        ; Debug "Settings loaded..."
      EndIf
      
    EndIf
    
    *ChildNode = NextXMLNode(*ChildNode)
  Wend
  
  If XML_Settings\ScreenWidth = 0 Or XML_Settings\ScreenHeight = 0
    ; use desktop settings
    ExamineDesktops()
    XML_Settings\ScreenWidth = DesktopWidth(0)
    XML_Settings\ScreenHeight = DesktopHeight(0)
  EndIf
  
  XML_CurrentPageNode = ChildXMLNode(XML_MainNode)
    
  ProcedureReturn #True
  
EndProcedure

; load the next XML page of the current loaded XML
Procedure.b LoadNextXMLPage()
  Repeat 
    If XML_CurrentPageNode.l <> 0
      ; next one
      XML_CurrentPageNode.l = NextXMLNode(XML_CurrentPageNode.l)
      If XML_CurrentPageNode.l = 0
        ProcedureReturn #False; no more nodes (finish presentation)
      EndIf
    Else
      ; first one
      XML_CurrentPageNode.l = ChildXMLNode(XML_MainNode)
    EndIf
    
    NodeName.s = ""
    If XMLNodeType(XML_CurrentPageNode.l) = #PB_XML_Normal
      NodeName.s = GetXMLNodeName(XML_CurrentPageNode.l)
    EndIf
  
  Until NodeName.s = "page" Or (NodeName.s = "" And XMLNodeType(XML_CurrentPageNode.l) <> #PB_XML_Comment)
    
  Select NodeName.s
    Case "page"
      ;Debug "New Page"
      InsertXMLElements(XML_CurrentPageNode.l)
      
    Case "settings"
      ; Debug "ignore settings node..."
      
    Case ""
      ; nothing found
      ProcedureReturn #False; no more nodes (finish presentation)
      
  EndSelect
    
  ProcedureReturn #True

EndProcedure

; load the previous XML page of the current loaded XML
Procedure.b LoadPreviousXMLPage()
  
  Repeat
    If XML_CurrentPageNode.l <> 0
      ; previous one
      XML_CurrentPageNode.l = PreviousXMLNode(XML_CurrentPageNode.l)
      If XML_CurrentPageNode.l = 0
        XML_CurrentPageNode.l = ChildXMLNode(XML_MainNode)
        ProcedureReturn #True; no more nodes (first one)
      EndIf
    Else
      ; first one
      XML_CurrentPageNode.l = ChildXMLNode(XML_MainNode)
    EndIf
    NodeName.s = ""
    If XMLNodeType(XML_CurrentPageNode.l) = #PB_XML_Normal
      NodeName.s = GetXMLNodeName(XML_CurrentPageNode.l)
    EndIf
  Until NodeName.s = "page" Or (NodeName.s = "" And XMLNodeType(XML_CurrentPageNode.l) <> #PB_XML_Comment)
  
  Select NodeName.s
    Case "page"
      ;Debug "New Page"
      InsertXMLElements(XML_CurrentPageNode.l)
      
    Case "settings"
      ; Debug "ignore settings node..."
      
    Case ""
      ; nothing found
      ProcedureReturn #False; no more nodes (finish presentation)
      
  EndSelect
    
  ProcedureReturn #True
  
EndProcedure

; ---------------------------------------------------------------
; Init
; ---------------------------------------------------------------

If CountProgramParameters()
  Filename$ = ProgramParameter(0)
Else
  Filename$ = "slide.xml"
EndIf

; load xml definition
If OpenSlideXML(Filename$) = #False
  Debug "Can not load XML file"
;  End
EndIf

; ---------------------------------------------------------------
; Screen
; ---------------------------------------------------------------
If XML_Settings\ScreenFull = 0
  ; windowed
  ScreenRet.l = OpenDisplayWindow()
Else
  ; fullscreen-mode
  ScreenRet.l = OpenDisplayScreen()
EndIf

If ScreenRet = 0
  ; screen successfully opened?
  MessageRequester("error", "Can not open screen! Please check your XML settings values...", #MB_ICONERROR)
  End
EndIf

SetFrameRate(60); max 60 frames

; load first page
If LoadNextXMLPage() = #True
  ; init new page
  PageCounter.l = 1
Else
  Debug "No valid page information found in XML definition file!"
  End
EndIf

; define the center of the screen (perspective)
SetPerspective(XML_Settings\ScreenWidth / 2, XML_Settings\ScreenHeight / 2)

; ---------------------------------------------------------------
; Main Loop
; ---------------------------------------------------------------
PageStartTime.l = ElapsedMilliseconds()
LastMousePush.l = ElapsedMilliseconds()
Quit = #False


Repeat
  
  MBR.l = 0
  MBL.l = 0
  
  ; manage events
  If IsFullscreen.b = #False
    ; check only in windowed mode
    Repeat
      Event = WindowEvent()
            
      Select Event
          
        Case #WM_RBUTTONDOWN
          MBR = 1
        Case #WM_LBUTTONDOWN
          MBL = 1
        Case #PB_Event_Menu
          Select EventMenu()
            Case #MenuOpen
              XMLFile$ = OpenFileRequester("Open", GetPathPart(ProgramFilename()), "XML (*.xml)|*.xml", 0 )
              If XMLFile$
                If OpenSlideXML(XMLFile$)
                  ; load first page
                  If LoadNextXMLPage() = #True
                    ; init new page
                    PageCounter.l = 1
                  Else
                    Debug "No valid page information found in XML definition file!"
                    End
                  EndIf
                EndIf
              EndIf
            Case #MenuExit
              Quit = #True
            Case #MenuAbout
              About$ = "        Pres3D from Kukulkan" + #LF$
              About$ + #LF$
              About$ + "kindly given to the PB community" + #LF$
              About$ + #LF$
              About$ + "Extended by:" + #LF$
              About$ + #LF$
              About$ + "Vera" + #LF$
              About$ + "Infratec" + #LF$
              MessageRequester("Pres3D V" + #Version, About$)
          EndSelect
        Case #PB_Event_CloseWindow
          Quit = #True
      EndSelect
    Until Event = 0
  EndIf
  
  ; draw
  ClearScreen(XML_Settings\BackgroundColor)
  ExamineKeyboard()
  ExamineMouse()
  ExamineKeyboard()
  
  ; manage mouse
  Zoom.l = Zoom.l + MouseWheel() * 20
  ; Debug "current zoom factor: " + Str(Zoom.l)
  
  If KeyboardPushed(#PB_Key_Return)
    
    ; clear keyboard input buffer until key is released
    While KeyboardPushed(#PB_Key_Return)
      ExamineKeyboard()
    Wend
    
    If IsFullscreen.b = #True
      OpenDisplayWindow()
    Else
      OpenDisplayScreen()
    EndIf
    
    ; reload current xml file
    OpenSlideXML(XML_CurrentXMLFile.s)
    If LoadNextXMLPage() = #True
      ; init new page
      PageCounter.l = 1
    Else
      Debug "Error reloading XML file " + XML_CurrentXMLFile.s
    EndIf
    
  EndIf
   
  If Start3D() = 0: Debug "error initializing 3d": End : EndIf
  
  CompilerIf #PB_Compiler_OS = #PB_OS_Windows
    !extrn _PB_Screen_Direct3DDevice
    !MOV dword EAX, [_PB_Screen_Direct3DDevice]
    !MOV dword [v_pd3d],EAX
    pd3d\SetRenderState(22,1)
  CompilerEndIf
  
  ; draw all ingoing effects
  TransformOngoing.b = #False; used to check, if any animation is not finished (no page-jump possible until this is #False)
  For x.l = 1 To InObjectCount.l
    With InObjects(x.l)
      If ElapsedMilliseconds() - PageStartTime.l > \Delay 
        If \Transform = #True
          *In.structEffectParameters = \EffectIn
          EffectTransformIn(*In, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, x.l, InObjectCount.l)
          TransformOngoing.b = #True ; some objects are still in progress
          ; transform object with new values
          SetSuperSpritePosition(\SpriteID, \ActualPosX, \ActualPosY, \ActualPosZ + Zoom.l)
          RotateSuperSprite3D(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
        EndIf
        If \ActualPosZ + Zoom.l > 0
          If \Type = 0 And XML_Settings\Smoothing = #True
            Sprite3DQuality(#PB_Sprite3D_BilinearFiltering) ; text
          Else
            Sprite3DQuality(#PB_Sprite3D_NoFiltering) ; image
          EndIf
          ; draw object
          DisplaySuperSprite3D(\SpriteID, \ActualTransparency)
        EndIf
      Else
        TransformOngoing.b = #True ; some objectas are still waiting!
      EndIf
    EndWith
  Next ; x.l
  
  ; draw all outgoing effects
  For x.l = 1 To OutObjectCount.l
    With OutObjects(x.l)
      If \Transform = #True
        TransformOngoing.b = #True ; some objects are still in progress
        If ElapsedMilliseconds() - PageStartTime.l > \Delay 
          *Out.structEffectParameters = \EffectOut
          EffectTransformOut(*Out, XML_Settings\ScreenWidth, XML_Settings\ScreenHeight, x.l, OutObjectCount.l)
          ; transform object with new values
          SetSuperSpritePosition(\SpriteID, \ActualPosX, \ActualPosY, \ActualPosZ + Zoom.l)
          RotateSuperSprite3D(\SpriteID, \ActualAngleX, \ActualAngleY, \ActualAngleZ)
        EndIf
      EndIf
      If \ActualPosZ + Zoom.l > 0
        If \Type = 0 And XML_Settings\Smoothing = #True
          Sprite3DQuality(#PB_Sprite3D_BilinearFiltering) ; text
        Else
          Sprite3DQuality(#PB_Sprite3D_NoFiltering) ; image
        EndIf
        ; draw object
        DisplaySuperSprite3D(\SpriteID, \ActualTransparency)
      EndIf
    EndWith
  Next ; x.l
  
  If IsFullscreen.b = #True
    MBL = MouseButton(#PB_MouseButton_Left)  ; forward
    MBR = MouseButton(#PB_MouseButton_Right) ; back
  EndIf
  
  If KeyboardPushed(#PB_Key_Left)     : MBR = 1: EndIf ; back
  If KeyboardPushed(#PB_Key_Right)    : MBL = 1: EndIf ; forward
  If KeyboardPushed(#PB_Key_PageUp)   : MBR = 1: EndIf ; back
  If KeyboardPushed(#PB_Key_PageDown) : MBL = 1: EndIf ; forward
  If KeyboardPushed(#PB_Key_Space)    : MBL = 1: EndIf ; forward
  
  
  ; page management
  If TransformOngoing.b = #False And (MBL.l <> 0 Or (MBR.l <> 0 And PageCounter.l > 1))
    ; ready with current transforming
    LoadSuccess.b = #False
    If MBL.l <> 0
      ; left mousebutton pressed (forward)
      PageCounter + 1; initialize next page
      ActivateOutgoing() ; activate outgoing effects of current elements
      ; load new page and start ingoing effects
      If LastPageReached.b = #False: LoadSuccess.b = LoadNextXMLPage(): EndIf
      PageStartTime.l = ElapsedMilliseconds()
    Else
      ; right mousebutton pressed (backward)
      PageCounter - 1; initialize previous page
      ActivateOutgoing() ; activate outgoing effects of current elements
      ; load new page and start ingoing effects
      If LastPageReached.b = #False: LoadSuccess.b = LoadPreviousXMLPage(): EndIf
      PageStartTime.l = ElapsedMilliseconds()
    EndIf

    If LoadSuccess.b = #True
      ; successfully initialized new page
    Else
      ; no success with new page
      If LastPageReached.b = #False
        ; Debug "No more page information found!"
        LastPageReached.b = #True
      Else
        Quit = #True
      EndIf
      
    EndIf
  EndIf
  
  Stop3D()
  
  If LastPageReached.b = #True
    ; show "last page" information
    StartDrawing(ScreenOutput())
    DrawText(1,1, "finished", RGB(100,110,128), RGB(255,255,255))
    StopDrawing()
  EndIf
  
  FlipBuffers()
  Delay(1) ; free some system ressources (keep CPU load below a critical value)

Until Quit = #True Or KeyboardPushed(#PB_Key_Escape)

FreeXML(#XML)

End

; IDE Options = PureBasic 4.50 (Windows - x86)
; CursorPosition = 8
; Folding = AAA-
; EnableAsm
; EnableXP
; Executable = Pres3D.exe
; DisableDebugger
; CompileSourceDirectory